import os
import random
import string
import subprocess
import tempfile
from enum import Enum
from pathlib import Path
from typing import Dict, Any, Optional

import orjson

from .dump import APIDump

_temp_folder = Path(tempfile.gettempdir())
_seed_letters = list(string.ascii_letters)


class VersionType(Enum):
    windows = "windows"
    """A Windows studio version.
    It contains the executable file (RobloxStudioBeta.exe) and is commonly located in %LocalAppData%/Roblox/Versions.
    """
    macos = "macos"
    """A macOS studio version.
    It should be an .app folder and contain a folder "Contents" with a file called "Info.plist".
    It is commonly located in ~/Applications."""


class Version:
    def __init__(self, path: Path, version_type: VersionType):
        self.path: Path = path
        self.version_type: VersionType = version_type

    @property
    def app_settings_file_path(self):
        return self._root_resources_path / "AppSettings.xml"

    @property
    def _root_resources_path(self):
        if self.version_type == VersionType.macos:
            return self.path / "Contents" / "Resources"
        else:
            return self.path

    @property
    def _root_executables_path(self):
        if self.version_type == VersionType.macos:
            return self.path / "Contents" / "MacOS"
        else:
            return self.path

    @property
    def content_folder_path(self):
        return self._root_resources_path / "content"

    @property
    def extra_content_folder_path(self):
        return self._root_resources_path / "ExtraContent"

    @property
    def platform_content_folder_path(self):
        return self._root_resources_path / "PlatformContent"

    @property
    def built_in_plugins_folder_path(self):
        return self._root_resources_path / "BuiltInPlugins"

    @property
    def built_in_standalone_plugins_folder_path(self):
        return self._root_resources_path / "BuiltInStandalonePlugins"

    @property
    def plugins_folder_path(self):
        return self._root_resources_path / "Plugins"

    @property
    def qml_folder_path(self):
        return self.path / "Qml"

    @property
    def shaders_folder_path(self):
        return self._root_resources_path / "shaders"

    @property
    def ssl_folder_path(self):
        return self._root_resources_path / "ssl"

    @property
    def cacert_file_path(self):
        return self.ssl_folder_path / "cacert.pem"

    @property
    def studio_fonts_folder_path(self):
        return self._root_resources_path / "StudioFonts"

    @property
    def client_settings_folder_path(self):
        return self._root_executables_path / "ClientSettings"

    @property
    def client_app_settings_file_path(self):
        return self.client_settings_folder_path / "ClientAppSettings.json"

    @property
    def reflection_metadata_file_path(self):
        return self._root_resources_path / "ReflectionMetadata.xml"

    @property
    def ribbon_file_path(self):
        return self._root_resources_path / "RobloxStudioRibbon.xml"

    @property
    def binary_file_path(self):
        if self.version_type == VersionType.macos:
            return self._root_executables_path / "RobloxStudio"
        else:
            return self._root_resources_path / "RobloxStudioBeta.exe"

    def get_fflag_overrides(self) -> Dict[str, Any]:
        """
        Gets the active FFlag overrides for this Studio version.
        If no overrides have been set, an empty dictionary is returned.
        """

        try:
            with open(
                    file=self.client_app_settings_file_path,
                    mode="rb"
            ) as client_app_settings_file:
                data = client_app_settings_file.read()

            return orjson.loads(data)
        except FileNotFoundError:
            return {}

    def set_fflag_overrides(self, overrides: Dict[str, Any]):
        """
        Sets the FFlag overrides for this Studio version.
        """

        try:
            os.mkdir(self.client_settings_folder_path)
        except FileExistsError:
            pass

        fflag_overrides_json = orjson.dumps(overrides)

        with open(
                file=self.client_app_settings_file_path,
                mode="wb"
        ) as client_app_settings_file:
            client_app_settings_file.write(fflag_overrides_json)

    def save_api_dump_to_path(self, path: Path):
        """
        Generates an API dump for this Roblox Studio version and places it in the specified path.
        """

        return subprocess.run(
            args=[
                self.binary_file_path, "-API", str(path)
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )

    def generate_api_dump_json(self, temp_path: Optional[Path] = None) -> dict:
        """
        Generates an API dump for this Roblox Studio version and returns its raw JSON representation.
        If temp_path is not specified, the path is generated randomly in a temporary directory and subsequently deleted.
        If temp_path is specified, you are expected to handle the deletion of the file yourself.
        """
        needs_deletion = False
        if not temp_path:
            random_seed = [random.choice(_seed_letters) for _ in range(8)]
            temp_path = _temp_folder / f"{random_seed}_dump.json"
            needs_deletion = True

        try:
            self.save_api_dump_to_path(temp_path)

            with open(temp_path, "rb") as file:
                dump_json = file.read()

            return orjson.loads(dump_json)
        finally:
            if needs_deletion:
                try:
                    os.remove(temp_path)
                except FileNotFoundError:
                    pass

    def generate_api_dump(self, temp_path: Optional[Path] = None) -> APIDump:
        """
        Generates an API dump for this Roblox Studio version and parses it.
        If temp_path is not specified, the path is generated randomly in a temporary directory and subsequently deleted.
        If temp_path is specified, you are expected to handle the deletion of the file yourself.
        """

        return APIDump(**self.generate_api_dump_json(temp_path=temp_path))

    def launch(
            self,
            *,
            file: Optional[Path] = None,
            base_url: Optional[str] = None,
            disable_user_plugins: bool = False,
    ):
        """
        Launches Roblox Studio and returns its launch process.

        Arguments:
            file: The file to launch Studio with, like a rbxl file.
            base_url: A base URL to use when sending requests. Default is "roblox.com".
            disable_user_plugins: Disables the loading of all local or remote user plugins.
        """

        args = [self.binary_file_path]

        if file:
            args.append(file)

        if base_url:
            args.append("-baseUrl")
            args.append(base_url)

        if disable_user_plugins:
            args.append("-disableLoadUserPlugins")

        process = subprocess.Popen(
            args=args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )

        return process

    def resolve_rbxasset_to_path(self, path: str) -> Optional[Path]:
        """
        Resolves a rbxasset path, like /textures/face.png.
        This function does not ensure that the path does not escape the directory - do not call this function with
        unfiltered user input.
        """
        for base_path in [
            self.platform_content_folder_path / "pc",
            self.extra_content_folder_path,
            self.content_folder_path
        ]:
            file_path = base_path / path
            if file_path.exists():
                return file_path
        return None



import os
from pathlib import Path
from typing import Dict, Any, Optional

import orjson

from .cookie import RobloxCookie
from .environments import Version, VersionType
from .settings import get_raw_settings
from .storage import AppStorage

_is_windows = os.name == "nt"
_is_posix = os.name == "posix"


class StudioClient:
    def __init__(
            self,
            roblox_path: Path = None,
    ):
        if roblox_path is None:
            if os.name == "nt":
                roblox_path = Path(os.getenv("LocalAppData")) / "Roblox"
            else:
                roblox_path = Path("~/Library").expanduser() / "Roblox"

        self.root_path: Path = roblox_path

    @property
    def global_settings_file_path(self):
        return self.root_path / "GlobalSettings_13.xml"

    @property
    def global_basic_settings_file_path(self):
        return self.root_path / "GlobalBasicSettings_13.xml"

    @property
    def analystics_settings_file_path(self):
        return self.root_path / "AnalysticsSettings.xml"

    @property
    def client_settings_folder_path(self):
        return self.root_path / "ClientSettings"

    @property
    def studio_app_settings_file_path(self):
        return self.client_settings_folder_path / "StudioAppSettings.json"

    @property
    def local_storage_folder_path(self):
        return self.root_path / "LocalStorage"

    @property
    def logs_folder_path(self):
        if os.name == "nt":
            return self.root_path / "logs"
        else:
            return Path("~/Library").expanduser() / "Logs" / "Roblox"

    @property
    def app_settings_file_path(self):
        return self.local_storage_folder_path / "appStorage.json"

    def get_cached_fflags(self) -> Dict[str, Any]:
        with open(
                file=self.studio_app_settings_file_path,
                mode="rb"
        ) as file:
            data = file.read()
        return orjson.loads(data)

    def get_app_storage(self):
        with open(
                file=self.app_settings_file_path,
                mode="rb"
        ) as file:
            data = file.read()
        return AppStorage(orjson.loads(data))

    def get_version(self, path: Path, version_type: Optional[VersionType] = None) -> Version:
        if version_type is None:
            if _is_windows:
                version_type = VersionType.windows
            elif _is_posix:
                version_type = VersionType.macos
            else:
                raise Exception("Couldn't determine your OS type. Specify the version_type parameter to silence this.")

        return Version(
            path=path,
            version_type=version_type
        )

    def get_cookies(self) -> Dict[str, Dict[str, RobloxCookie]]:
        """
        Gets the stored cookies for the current Roblox session.

        Returns:
            A dictionary where the keys are cookie hosts (e.g. "roblox.com") and the values are dictionaries where the
            keys are the names of the cookie (e.g. ".ROBLOSECURITY") and the values are the RobloxCookie objects themselves.

            Example: {"roblox.com": {".ROBLOSECURITY": RobloxCookie}}
        """
        cookie_dict = {}
        raw_cookie_dict = get_raw_settings("RobloxStudioBrowser")

        for url, raw_cookies in raw_cookie_dict.items():
            new_cookies = {}
            for cookie_name, raw_cookie_value in raw_cookies.items():
                new_cookies[cookie_name] = RobloxCookie.from_raw_string(raw_cookie_value)
            cookie_dict[url] = new_cookies

        return cookie_dict

    def set_cookies(self, cookies: Dict[str, Dict[str, RobloxCookie]]):
        raw_cookie_dict = {}

        for url, url_cookies in cookies.items():
            raw_cookie_dict[url] = {cookie_name: cookie.to_raw_string() for cookie_name, cookie in url_cookies.items()}

        return raw_cookie_dict



import datetime

from pydantic import BaseModel


def raw_cookie_to_dict(string: str) -> dict:
    result = {}
    string_pairs = string.split(",")
    for string_pair in string_pairs:
        raw_key, enclosed_raw_value = string_pair.strip().split("::")
        raw_value = enclosed_raw_value[1:-1]
        result[raw_key] = raw_value

    return result


def dict_to_raw_cookie(cookie_dict: dict) -> str:
    return ",".join(f"{raw_key}::<{raw_value}>" for raw_key, raw_value in cookie_dict.items())


class RobloxCookie(BaseModel):
    secure: bool
    expires: datetime.datetime
    value: str

    @classmethod
    def from_raw_string(cls, string: str):
        decoded_pairs = raw_cookie_to_dict(string)

        return cls(
            secure=True if decoded_pairs.get("SEC") and decoded_pairs["SEC"].lower() == "yes" else False,
            expires=decoded_pairs["EXP"],
            value=decoded_pairs["COOK"]
        )

    def to_raw_string(self):
        return dict_to_raw_cookie({
            "SEC": "YES" if self.secure else "NO",
            "EXP": self.expires.isoformat(),
            "COOK": self.value
        })


if __name__ == '__main__':
    cookie_string = "SEC::<YES>,EXP::<2052-01-27T20:58:25Z>,COOK::<TEST>"
    cookie = RobloxCookie.from_raw_string(cookie_string)
    print(cookie)



from enum import Enum
from typing import List, Optional, Literal, Union

from pydantic import BaseModel, Field


class ThreadSafety(Enum):
    read_safe = "ReadSafe"
    unsafe = "Unsafe"
    safe = "Safe"


class SecurityLevel(Enum):
    none = "None"
    roblox_place_security = "RobloxPlaceSecurity"
    plugin_security = "PluginSecurity"
    local_user_security = "LocalUserSecurity"
    roblox_script_security = "RobloxScriptSecurity"
    roblox_security = "RobloxSecurity"
    not_accessible_security = "NotAccessibleSecurity"


class ClassTag(Enum):
    not_creatable = "NotCreatable"
    not_browsable = "NotBrowsable"
    service = "Service"
    not_replicated = "NotReplicated"
    deprecated = "Deprecated"
    player_replicated = "PlayerReplicated"
    settings = "Settings"
    user_settings = "UserSettings"


class MemoryCategory(Enum):
    instances = "Instances"
    animation = "Animation"
    physics_parts = "PhysicsParts"
    graphics_texture = "GraphicsTexture"
    gui = "Gui"
    script = "Script"
    internal = "Internal"


class MemberSerialization(BaseModel):
    can_load: bool = Field(alias="CanLoad")
    can_save: bool = Field(alias="CanSave")


class MemberSecurity(BaseModel):
    read: SecurityLevel = Field(alias="Read")
    write: SecurityLevel = Field(alias="Write")


class ValueType(BaseModel):
    category: str = Field(alias="Category")
    name: str = Field(alias="Name")


class ClassMemberProperty(BaseModel):
    member_type: Literal["Property"] = Field(alias="MemberType")
    name: str = Field(alias="Name")
    category: str = Field(alias="Category")
    security: MemberSecurity = Field(alias="Security")
    serialization: MemberSerialization = Field(alias="Serialization")
    thread_safety: ThreadSafety = Field(alias="ThreadSafety")
    value_type: ValueType = Field(alias="ValueType")


class FunctionParameter(BaseModel):
    name: str = Field(alias="Name")
    type: ValueType = Field(alias="Type")


class ClassMemberFunction(BaseModel):
    member_type: Literal["Function"] = Field(alias="MemberType")
    name: str = Field(alias="Name")
    parameters: List[FunctionParameter] = Field(alias="Parameters")
    return_type: ValueType = Field(alias="ReturnType")
    security: SecurityLevel = Field(alias="Security")
    thread_safety: ThreadSafety = Field(alias="ThreadSafety")


class ClassMemberEvent(BaseModel):
    member_type: Literal["Event"] = Field(alias="MemberType")
    name: str = Field(alias="Name")
    parameters: List[FunctionParameter] = Field(alias="Parameters")
    security: SecurityLevel = Field(alias="Security")
    thread_safety: ThreadSafety = Field(alias="ThreadSafety")


class ClassMemberCallback(BaseModel):
    member_type: Literal["Callback"] = Field(alias="MemberType")
    name: str = Field(alias="Name")
    parameters: List[FunctionParameter] = Field(alias="Parameters")
    return_type: ValueType = Field(alias="ReturnType")
    security: SecurityLevel = Field(alias="Security")
    thread_safety: ThreadSafety = Field(alias="ThreadSafety")


class APIDumpClass(BaseModel):
    members: List[Union[ClassMemberProperty, ClassMemberFunction, ClassMemberEvent, ClassMemberCallback]] = Field(
        alias="Members")
    tags: Optional[List[ClassTag]] = Field(None, alias="Tags")
    superclass: str = Field(alias="Superclass")
    name: str = Field(alias="Name")
    memory_category: MemoryCategory = Field(alias="MemoryCategory")


class APIDumpEnumItem(BaseModel):
    name: str = Field(alias="Name")
    value: int = Field(alias="Value")


class APIDumpEnum(BaseModel):
    items: List[APIDumpEnumItem] = Field(alias="Items")
    name: str = Field(alias="Name")


class APIDump(BaseModel):
    classes: List[APIDumpClass] = Field(alias="Classes")
    enums: List[APIDumpEnum] = Field(alias="Enums")
    version: int = Field(alias="Version")



from yarl import URL
from enum import Enum


class RobloxBranch(Enum):
    production = "production"

    sitetest1 = "sitetest1"
    sitetest2 = "sitetest2"
    sitetest3 = "sitetest3"

    gametest1 = "gametest1"
    gametest2 = "gametest2"


roblox_branch_to_url = {
    RobloxBranch.production: URL("https://setup.rbxcdn.com/"),

    RobloxBranch.sitetest1: URL("http://setup.sitetest1.robloxlabs.com.s3.amazonaws.com/"),
    RobloxBranch.sitetest2: URL("http://setup.sitetest2.robloxlabs.com.s3.amazonaws.com/"),
    RobloxBranch.sitetest3: URL("http://setup.sitetest3.robloxlabs.com.s3.amazonaws.com/"),

    RobloxBranch.gametest1: URL("http://setup.gametest1.robloxlabs.com.s3.amazonaws.com/"),
    RobloxBranch.gametest2: URL("http://setup.gametest2.robloxlabs.com.s3.amazonaws.com/")
}



import os
from pathlib import Path

if os.name == "nt":
    import winreg
elif os.name == "posix":
    import plistlib


def get_subkeys(key):
    index = 0

    while True:
        try:
            yield winreg.EnumKey(key, index)
        except OSError:
            break
        index += 1


def get_key_values(key):
    index = 0

    while True:
        try:
            yield winreg.EnumValue(key, index)
        except OSError:
            break
        index += 1


def key_to_dictionary(key):
    result = {}

    for value_name, value_data, value_type in get_key_values(key):
        result[value_name] = value_data

    for subkey_name in get_subkeys(key):
        subkey = winreg.OpenKeyEx(key, subkey_name)
        result[subkey_name] = key_to_dictionary(subkey)

    return result


def deep_rbx_dict_to_list(data):
    """
    Deeply converts "rbx dicts" (ie, dicts with ordered numeric keys) into proper lists.
    May not maintain order.
    """

    bad_dict: dict = data

    for key, value in bad_dict.items():
        if not isinstance(value, dict):
            continue

        bad_dict[key] = deep_rbx_dict_to_list(value)

    bad_keys = set(bad_dict.keys())

    if "size" not in bad_keys:
        return bad_dict

    numeric_keys = set()

    for bad_key in bad_keys:
        if bad_key.isnumeric():
            numeric_keys.add(int(bad_key))
        else:
            if bad_key != "size":
                # if there is a key in the dict other than size that is not numeric, this is not a proper list
                return bad_dict

    list_size = 0
    for bad_key in bad_keys:
        if bad_key.isnumeric():
            index = int(bad_key)
            if index > list_size:
                list_size = index

    new_list = [None for _ in range(list_size)]

    for bad_key in bad_keys:
        if bad_key.isnumeric():
            new_list[int(bad_key) - 1] = bad_dict[bad_key]

    return new_list


def plist_data_to_rbx_dict(plist_data: dict) -> dict:
    """
    Converts data from a Roblox macOS .plist file into a dictionary.
    """
    rbx_dict: dict = {}
    for raw_key, raw_value in plist_data.items():
        split_key = raw_key.split(".")

        key_piece = rbx_dict

        for i, key_bit in enumerate(split_key):
            key_bit = key_bit.replace("\u00b7", ".")
            if i == len(split_key) - 1:
                if isinstance(raw_value, bytes):
                    raw_value = str(raw_value)
                key_piece[key_bit] = raw_value
                continue

            if key_piece.get(key_bit):
                key_piece = key_piece[key_bit]
            else:
                new_piece = {}
                key_piece[key_bit] = new_piece
                key_piece = new_piece
    return rbx_dict


def get_raw_settings(name: str):
    if os.name == "nt":
        key = winreg.OpenKeyEx(
            key=winreg.HKEY_CURRENT_USER,
            sub_key=rf"SOFTWARE\Roblox\{name}"
        )

        settings = key_to_dictionary(key)
        settings = deep_rbx_dict_to_list(settings)

        key.Close()
    elif os.name == "posix":
        plist_path = Path("~/Library/Preferences").expanduser() / f"com.roblox.{name}.plist"

        with open(
                file=plist_path,
                mode="rb"
        ) as roblox_plist_file:
            plist_data = plistlib.load(roblox_plist_file)
        settings = plist_data_to_rbx_dict(plist_data)
        settings = deep_rbx_dict_to_list(settings)
    else:
        raise NotImplementedError("Unknown OS")

    return settings


if __name__ == '__main__':
    print(get_raw_settings("RobloxStudio"))



from __future__ import annotations

# import warnings

from datetime import datetime
from enum import Enum
from re import compile
from typing import List, Optional, Union, Tuple

from yarl import URL
from aiohttp import ClientSession
from dateutil.parser import parse

from .branches import RobloxBranch, roblox_branch_to_url
from .dump import APIDump

cdn_url = URL("https://setup.rbxcdn.com/")
mac_cdn_url = cdn_url / "mac"

git_hash_pattern = compile(r"New ([^ ]*?) (version-[^ ]*) at ([^ ]*) (.*?), file version: ([0123456789, ]*), "
                           r"git hash: ?([^ ]*)")
file_version_pattern = compile(r"New ([^ ]*?) (version-[^ ]*) at ([^ ]*) (.*?), file vers?ion: ?([0123456789, ]*)")
fallback_pattern = compile(r"New ([^ ]*?) (version-[^ ]*) at ([^ ]*) (.*)")

reverting_pattern = compile(r"Reverting ([^ ]*?) to version (version-[^ ]*) at ([^ ]*) (.*)")
revert_pattern = compile(r"Revert ([^ ]*?) (version-[^ ]*) at ([^ ]*) (.*)")


class OperatingSystem(Enum):
    windows = "windows"
    mac = "mac"


class BinaryType(Enum):
    windows_player = "WindowsPlayer"
    windows_studio = "WindowsStudio"
    mac_player = "MacPlayer"
    mac_studio = "MacStudio"


class DeploymentType(Enum):
    rcc_service = "rcc_service"

    client = "client"
    windows_player = "windows_player"

    studio = "studio"
    studio_64 = "studio_64"
    studio_beta = "studio_beta"
    mfc_studio = "mfc_studio"

    windows_mfc_player_and_studio = "windows_mfc_player_and_studio"


_log_name_to_deployment_type = {
    "RccService": DeploymentType.rcc_service,
    "Client": DeploymentType.client,
    "WindowsPlayer": DeploymentType.windows_player,

    "Studio": DeploymentType.studio,
    "Studio64": DeploymentType.studio_64,
    "StudioBeta": DeploymentType.studio_beta,
    "MFCStudio": DeploymentType.mfc_studio,

    "windows-mfc-player-and-studio": DeploymentType.windows_mfc_player_and_studio
}


class DeploymentPackage:
    def __init__(self, client: DeploymentClient, branch: RobloxBranch, deployment: Deployment, lines: List[str]):
        self._branch: RobloxBranch = branch
        self._client: DeploymentClient = client
        self._deployment: Deployment = deployment

        self.name: str = lines[0]
        self.md5: str = lines[1]
        self.compressed_size: int = int(lines[2])
        self.size: int = int(lines[3])

    @property
    def url(self):
        return cdn_url / f"{self._deployment.version_hash}-{self.name}"


class DeploymentPackages:
    def __init__(self, client: DeploymentClient, branch: RobloxBranch, deployment: Deployment, packages_data: str):
        self._branch: RobloxBranch = branch
        self._client: DeploymentClient = client
        self._deployment: Deployment = deployment

        self.packages: List[DeploymentPackage] = []

        packages_lines = packages_data.splitlines()
        packages_file_lines = packages_lines[1:]

        self.version: str = packages_lines[0]

        for i in range(0, len(packages_file_lines), 4):
            file_lines = packages_file_lines[i:i + 4]
            self.packages.append(DeploymentPackage(
                client=self._client,
                branch=self._branch,
                deployment=self._deployment,
                lines=file_lines
            ))


class Deployment:
    def __init__(
            self,
            client: DeploymentClient,
            branch: RobloxBranch,
            operating_system: OperatingSystem,
            history_line: str
    ):
        self._branch: RobloxBranch = branch
        self._client: DeploymentClient = client
        self._operating_system: OperatingSystem = operating_system

        self.deployment_type: DeploymentType
        self.version_hash: str
        self.timestamp: datetime
        self.version_number: Optional[Tuple[int, int, int, int]] = None
        self.git_hash: Optional[str] = None

        if "git hash" in history_line:
            match = git_hash_pattern.search(string=history_line)
            assert match

            self.deployment_type = _log_name_to_deployment_type.get(match.group(1))
            self.version_hash = match.group(2)
            date_string = match.group(3)
            time_string = match.group(4)
            self.timestamp = parse(f"{date_string} {time_string}")
            version_string = match.group(5).strip()
            if version_string:
                self.version_number = tuple(int(piece.strip()) for piece in version_string.split(","))
            self.git_hash = match.group(6)
        elif "file version" in history_line or "file verion" in history_line:
            match = file_version_pattern.search(string=history_line)
            assert match

            self.deployment_type = _log_name_to_deployment_type.get(match.group(1))
            self.version_hash = match.group(2)
            date_string = match.group(3)
            time_string = match.group(4)
            self.timestamp = parse(f"{date_string} {time_string}")
            version_string = match.group(5).strip()
            if version_string:
                self.version_number = tuple(int(piece.strip()) for piece in version_string.split(","))
        else:
            match = fallback_pattern.search(string=history_line)
            assert match

            self.deployment_type = _log_name_to_deployment_type.get(match.group(1))
            self.version_hash = match.group(2)
            date_string = match.group(3)
            time_string = match.group(4)
            self.timestamp = parse(f"{date_string} {time_string}")

        if self.version_number and self.version_number[0] >= 2000:
            self.version_number = None

    def get_url(self, item: str):
        branch_url = roblox_branch_to_url.get(self._branch)
        branch_os_url = branch_url / "mac" if self._operating_system == OperatingSystem.mac else branch_url
        return branch_os_url / f"{self.version_hash}-{item}"

    def get_root_zip(self):
        if self.deployment_type in {DeploymentType.studio, DeploymentType.studio_64}:
            return self.get_url("RobloxStudio.zip")
        elif self.deployment_type == DeploymentType.client:
            return self.get_url("RobloxApp.zip")
        else:
            raise NotImplementedError

    async def get_api_dump(self) -> APIDump:
        async with self._client.session.get(self.get_url("API-Dump.json")) as dump_response:
            dump_response.raise_for_status()
            return APIDump(**await dump_response.json())

    async def get_packages(self) -> DeploymentPackages:
        assert self._operating_system == OperatingSystem.windows, "Cannot get packages for non-Windows installs"

        async with self._client.session.get(self.get_url("rbxPkgManifest.txt")) as packages_response:
            packages_response.raise_for_status()
            return DeploymentPackages(
                client=self._client,
                branch=self._branch,
                deployment=self,
                packages_data=await packages_response.text(encoding="utf-8")
            )


class DeploymentRevert:
    def __init__(
            self,
            client: DeploymentClient,
            branch: RobloxBranch,
            operating_system: OperatingSystem,
            history_line: str
    ):
        self._branch: RobloxBranch = branch
        self._client: DeploymentClient = client
        self._operating_system: OperatingSystem = operating_system

        self.deployment_type: DeploymentType
        self.version_hash: str
        self.timestamp: datetime
        self.git_hash: Optional[str] = None

        if history_line.startswith("Reverting"):
            match = reverting_pattern.search(history_line)
            self.deployment_type = _log_name_to_deployment_type.get(match.group(1))
            self.version_hash = match.group(2)
            date_string = match.group(3)
            time_string = match.group(4)
            self.timestamp = parse(f"{date_string} {time_string}")
        else:
            match = revert_pattern.search(history_line)
            self.deployment_type = _log_name_to_deployment_type.get(match.group(1))
            self.version_hash = match.group(2)
            date_string = match.group(3)
            time_string = match.group(4)
            self.timestamp = parse(f"{date_string} {time_string}")


class DeploymentHistory:
    def __init__(
            self,
            client: DeploymentClient,
            branch: RobloxBranch,
            operating_system: OperatingSystem,
            history_data: str
    ):
        self._client: DeploymentClient = client
        self._branch: RobloxBranch = branch
        self._operating_system: OperatingSystem = operating_system

        self.history: List[Union[Deployment, DeploymentRevert]] = []

        history_split = history_data.splitlines()
        for history_line in history_split:
            history_line = history_line.strip()
            if not history_line:
                continue

            history_sublines = history_line.split("...")

            for history_subline in history_sublines:
                history_subline = history_subline.strip()

                if history_subline == "Done!" or history_subline == "Error!":
                    continue

                try:
                    if history_subline.startswith("New"):
                        self.history.append(Deployment(
                            client=self._client,
                            branch=self._branch,
                            operating_system=self._operating_system,
                            history_line=history_subline
                        ))
                    elif history_subline.startswith("Revert"):
                        self.history.append(DeploymentRevert(
                            client=self._client,
                            branch=self._branch,
                            operating_system=self._operating_system,
                            history_line=history_subline
                        ))
                except AssertionError:
                    pass
                    # warnings.warn(f"Failed to parse string {history_subline!r}")

    def get_latest_deployment(self, deployment_type: DeploymentType) -> Optional[Deployment]:
        for deployment in reversed(self.history):
            if deployment.deployment_type == deployment_type:
                return deployment
        return None


class DeploymentClient:
    def __init__(self):
        self.session = ClientSession(
            headers={
                "User-Agent": "Roblox/WinInet"
            }
        )

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.session.close()

    async def get_deployments(self, branch: RobloxBranch, operating_system: OperatingSystem):
        branch_url = roblox_branch_to_url.get(branch)
        branch_os_url = branch_url / "mac" if operating_system == OperatingSystem.mac else branch_url

        async with self.session.get(branch_os_url / "DeployHistory.txt") as history_response:
            history_response.raise_for_status()
            return DeploymentHistory(
                client=self,
                branch=branch,
                operating_system=operating_system,
                history_data=await history_response.text(encoding="utf-8")
            )



from enum import IntEnum, Enum
from typing import Optional, List

import orjson

from .utilities import int_or_none


class MembershipType(IntEnum):
    """
    Represents a Roblox membership type.
    Please see https://developer.roblox.com/en-us/api-reference/enum/MembershipType for more information.
    """
    none = 0
    builders_club = 1
    turbo_builders_club = 2
    outrageous_builders_club = 3
    premium = 4


class PolicyServiceHttpResponse:
    def __init__(self, data: dict):
        self.is_subject_to_china_policies: bool = data["isSubjectToChinaPolicies"]
        self.are_paid_random_items_restricted: bool = data["arePaidRandomItemsRestricted"]
        self.is_paid_item_trading_allowed: bool = data["isPaidItemTradingAllowed"]
        self.allowed_external_link_references: List[str] = data["allowedExternalLinkReferences"]


class Theme(Enum):
    dark = "dark"
    light = "light"


class AppStorage:
    def __init__(self, data: dict):
        self.app_installation_id: Optional[int] = int_or_none(data.get("AppInstallationId"))
        self.username: Optional[str] = data.get("Username")
        self.membership: Optional[MembershipType] = MembershipType(int(data["Membership"])) if data.get("Membership") \
            else None
        self.roblox_locale_id: Optional[str] = data.get("RobloxLocaleId")
        self.browser_tracker_id: Optional[int] = int_or_none(data.get("BrowserTrackerId"))

        self.web_login: Optional[int] = int_or_none(data.get("WebLogin"))
        # rethink where to do these orjson loads. this is probably bad
        self.policy_service_http_response: Optional[PolicyServiceHttpResponse] = \
            PolicyServiceHttpResponse(orjson.loads(data["PolicyServiceHttpResponse"])) \
                if data.get("PolicyServiceHttpResponse") else None

        self.player_exe_launch_time: Optional[int] = int_or_none(data.get("PlayerExeLaunchTime"))
        self.user_id: Optional[int] = int_or_none(data.get("UserId"))
        self.is_under_13: Optional[int] = (True if data["IsUnder13"] == "true" else "false") if data.get("IsUnder13") \
            else None
        self.display_name: Optional[str] = data.get("DisplayName")
        self.country_code: Optional[str] = data.get("CountryCode")
        self.game_locale_id: Optional[str] = data.get("GameLocaleId")
        self.authenticated_theme: Optional[Theme] = Theme(data["AuthenticatedTheme"]) \
            if data.get("AuthenticatedTheme") else None
        self.experience_menu_version: Optional[str] = data.get("ExperienceMenuVersion")
        self.native_close_lua_prompt_display_count: Optional[dict] = \
            orjson.loads(data["NativeCloseLuaPromptDisplayCount"]) if data.get("NativeCloseLuaPromptDisplayCount") \
                else None
